;===============================================================================
;  simple calculator (add/sub/mul/div/sqrt, percent key, memory add/sub)
;  with exchangable input/output interface (e.g. Keypad + LCD/LED/VFD or RS232)
;===============================================================================
;
;Controller: ATTiny2313 @ 8 MHz (internal R/C-oscillator)
;Low  Fuse Byte: 0xC4 -> 8 MHz internal R/C osc., BOD enabled
;High Fuse Byte: 0xDB -> BOD Level = 2.7 V
;
;Ports: see I/O module
;
;Version History:
;  0.1.0  2007-06-16  Arne Rossius
;    * version for testing math functions, RS232 I/O interface only
;  0.3.0  2007-06-18  Arne Rossius
;    * working version with LED display and RS232 input, no mul/div and memory
;  0.4.0  2007-06-29  Arne Rossius
;    * added multiplication, still some rounding errors
;  0.5.0  2007-06-24  Arne Rossius
;    * new multiplication routine with better rounding
;    * remove (-) sign if result is zero
;    * percent: add, sub, copy (missing: mul, div)
;  0.6.0  2007-07-02  Arne Rossius
;    * percent: added mul, div (depending on not yet implemented division)
;    * changed input to 4x6 matrix keypad
;  0.7.0  2007-07-08  Arne Rossius
;    * BUGFIX: stack leak in multiplication routine corrupts return address
;    * BUGFIX: trailing zeros after decimal point after rounding
;    * BUGFIX: minus sign not removed if result is only zero after rounding
;    * added division routine
;    * display sign directly in front of the number if not all digits are used
;    * merged some redundant code into new functions
;  0.8.0  2007-07-09  Arne Rossius
;    * some minor bugfixes
;    * added memory
;    * cleaned up several routines, added second flags register
;    * added new public function in I/O module: 'io_updatesymbols'
;  0.8.1  2007-09-13  Arne Rossius
;    * BUGFIX: lost sign of result when using "CE"
;  0.8.2  2007-12-02  Arne Rossius
;    * updated RS232 I/O-module to include all available functions
;  0.9  2007-12-02  Arne Rossius
;    * BUGFIX: rounding for display failed if a new digit was introduced (e.g. 99.9->100)
;    * changed all numbers to packed BCD to save RAM (fixes bug with DISPLEN=16)
;    * new keypad scanning routine to save Flash memory (needed for sqrt and 1/x)
;    * new keystroke comparision routine (for RS232) to save Flash memory
;    * new percentage add/sub routine to save Flash memory
;    * added two new number registers (numstart, numiter) for sqrt iteration
;    * square root and reciprocal value functions
;    * additional (7th) row of keys in 'io_led_keypad.asm' for new functions
;    * changed some function names to represent the used number (mostly num2)
;    * several small bugfixes
;  0.9.1  2007-12-20  Arne Rossius
;    * added alternate keyboard layout in "io_led_keypad.asm"

.include "../include/tn2313def.inc"

;R0-R15 = reserved for I/O module
.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	flags = R19
	.equ	num1neg = 0 ;number 1 is negative
	.equ	num2neg = 1 ;number 2 is negative
	.equ	num3neg = 2 ;number 3 is negative
	.equ	ManualInput = 3 ;user input is present
	.equ	AnyInput = 4 ;any input is present (user or memory recall)
	.equ	InputNonzero = 5 ;entered any of: 1 2 3 4 5 6 7 8 9 dot
	.equ	Error = 6 ;calculation returned an error
	.equ	ShiftRounded = 7 ;last right-shift rounded *up*
.def	flags2 = R20
	.equ	OP_ADD = 0 ;next operation will be addition
	.equ	OP_SUB = 1 ;next operation will be subtraction
	.equ	OP_MUL = 2 ;next operation will be multiplication
	.equ	OP_DIV = 3 ;next operation will be division
	.equ	MemNeg = 4 ;number in memory is negative
	.equ	MemZero = 5 ;number in memory is zero
.def	inplen = R21
.def	decimals = R22
.def	first = R23
.def	last = R24
.def	length = R25
;R27:R26 = X pointer (reserved for I/O module)
;R29:R28 = Y pointer
;R31:R30 = Z pointer

;Keys '0' to '9' = 0x00 to 0x09
.equ	KEY_ADD = 0x10
.equ	KEY_SUB = 0x11
.equ	KEY_MUL = 0x12
.equ	KEY_DIV = 0x13
.equ	KEY_PERCENT = 0x14
.equ	KEY_EQUAL = 0x15
.equ	KEY_NEG = 0x16
.equ	KEY_CE = 0x17
.equ	KEY_AC = 0x18
.equ	KEY_DOT = 0x19
;optional memory keys:
.equ	KEY_MEMADD = 0x20
.equ	KEY_MEMSUB = 0x21
.equ	KEY_MEMRCL = 0x22
.equ	KEY_MEMCLR = 0x23
;optional special function keys:
.equ	KEY_INV = 0x40
.equ	KEY_SQRT = 0x41

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	
;.include "io_rs232.asm"
;.include "io_led_keypad.asm"
.include "io_vfd_keypad.asm"
.include "math.asm"

;===============================================================================

.dseg
.org 0x60
	numstart: .byte DISPSIZE
	numiter: .byte DISPSIZE
	num1: .byte DISPSIZE
	num2: .byte DISPSIZE
	num3: .byte DISPSIZE
	mem: .byte DISPSIZE
	stack: .byte 16
	
;===============================================================================

.cseg
reset:
	;init stackpointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;initialize I/O devices
	rcall	io_init
	
	;load 0 to memory
	rcall	clearmem
	
init:
	;initialize registers and load 0 to num1/num2/num3
	ldi	flags, 1<<ManualInput ;prevent (-) getting lost if pressed first
	andi	flags2, 1<<MemZero | 1<<MemNeg ;memory wasn't changed
	ldi	inplen, 0
	ldi	decimals, 0
	rcall	clearnum1
	rcall	clearnum2
	rcall	clearnum3
	
;-------------------------------------------------------------------------------

update: ;copy num1 (result) to num2 (displayed) and round to fit display size
	rcall	copy1to2 ;copy num1 to num2
	
update_nocopy:
	rcall	copy2to3 ;backup original number
update_loop:
	;get first and last displayed digit and calculate length
	ldi	ZL, num2
	rcall	numfirst
	cpi	first, DISPSIZE*2 ;no (-) sign if result is zero
	brlo	PC+2
	andi	flags, ~(1<<num2neg)
	cpi	first, DISPSIZE ;display at least 1 digit before decimal point
	brlo	PC+2
	ldi	first, DISPSIZE-1
	rcall	num2last
	mov	length, last
	sub	length, first
	inc	length
	
	;round if length > DISPSIZE
	andi	flags, ~(1<<ShiftRounded)
	cpi	length, DISPSIZE+1
	brlo	update_round_end
	;round
	ldi	temp, DISPSIZE
	add	temp, first ;temp = first digit that isn't displayed
	mov	temp2, temp
	lsr	temp2
	ldi	ZL, num2
	add	ZL, temp2
	ld	temp2, Z
	sbrc	temp, 0
	swap	temp2
	sbrc	flags, ShiftRounded ;subtract 1 if rounded before
	subi	temp2, 0x10
	cpi	temp2, 0x50
	brlo	update_round_end
	;round up: add 1 to last displayed digit
	dec	temp
	rcall	math_inc2
	ori	flags, 1<<ShiftRounded
	mov	temp2, first
	ldi	ZL, num2 ;recalculate first digit
	rcall	numfirst
	cp	first, temp2
	brlo	update_loop ;if rounding introduced a new digit, round again
	mov	first, temp2 ;restore old 'first' value in case new value is right of dp
update_round_end:

	;remove trailing zeros from *displayed* fractional part of the result
	cpi	length, DISPSIZE+1 ;limit length to DISPSIZE digits
	brlo	PC+2
	ldi	length, DISPSIZE
	mov	last, first
	add	last, length
	dec	last
	mov	temp2, last
	lsr	temp2
	ldi	ZL, num2
	add	ZL, temp2
update_trailingzero:
	cpi	last, DISPSIZE
	brlo	update_trailingzero_end
	ld	temp, Z
	sbrc	last, 0
	swap	temp
	cpi	temp, 0x10
	brsh	update_trailingzero_end
	sbrs	last, 0
	dec	ZL
	dec	last
	dec	length
	rjmp	update_trailingzero
update_trailingzero_end:
	;remove (-) sign if *displayed* part of the result is zero
	cpi	length, 1
	brne	update_nonzero
	ldi	ZL, num2+(DISPSIZE/2-1)
	ld	temp, Z
	cpi	temp, 0x00 ;upper nibble always zero if length=1
	brne	update_nonzero
	andi	flags, ~(1<<num2neg)
update_nonzero:
	;update display
	rcall	io_update
	rcall	copy3to2 ;restore original number

;-------------------------------------------------------------------------------

waitkey:
	rcall	io_getkey
	cpi	temp, KEY_AC
	brne	PC+2
	rjmp	init
	sbrc	flags, Error ;lock all keys but 'AC' if an error occured
	rjmp	waitkey
	cpi	temp, KEY_DOT
	breq	dot
	cpi	temp, KEY_NEG
	breq	flip_sign
	cpi	temp, KEY_ADD
	breq	addition
	cpi	temp, KEY_SUB
	breq	subtraction
	cpi	temp, KEY_MUL
	breq	multiplication
	cpi	temp, KEY_DIV
	breq	division
	cpi	temp, KEY_EQUAL
	breq	equal
	cpi	temp, KEY_INV
	breq	invert
	cpi	temp, KEY_SQRT
	breq	sqrt
	cpi	temp, KEY_CE
	breq	ce
	cpi	temp, KEY_MEMCLR
	breq	memclr
	cpi	temp, KEY_MEMRCL
	breq	memrcl
	cpi	temp, KEY_MEMADD
	breq	memadd
	cpi	temp, KEY_MEMSUB
	breq	memsub
	cpi	temp, KEY_PERCENT
	brne	PC+2
	rjmp	percent
	cpi	temp, 10
	brsh	waitkey
	rjmp	digit
	
dot:
	;decimal point
	cpi	decimals, 0
	brne	waitkey ;number already has a decimal point
	ldi	decimals, 1
	ori	flags, 1<<AnyInput | 1<<ManualInput | 1<<InputNonzero
	cpi	inplen, 0
	brne	waitkey
	rcall	clearnum2
	ldi	inplen, 1
	rjmp	digit_end
	
flip_sign:
	;flip sign
	sbrs	flags, AnyInput
	rjmp	flip_sign_result ;no input yet -> flip sign of last result
	ldi	temp, 1<<num2neg
	eor	flags, temp
	rjmp	update_nocopy
flip_sign_result:
	ldi	temp, 1<<num1neg
	eor	flags, temp
	rjmp	update

addition:
	;perform last requested operation, next operation will be addition
	rcall	perform_operation
	ori	flags2, 1<<OP_ADD
	rjmp	update
	
subtraction:
	;perform last requested operation, next operation will be subtraction
	rcall	perform_operation
	ori	flags2, 1<<OP_SUB
	rjmp	update
	
multiplication:
	;perform last requested operation, next operation will be multiplication
	rcall	perform_operation
	ori	flags2, 1<<OP_MUL
	rjmp	update
	
division:
	;perform last requested operation, next operation will be division
	rcall	perform_operation
	ori	flags2, 1<<OP_DIV
	rjmp	update
	
equal:
	;perform last requested operation, next operation will be num2->num1
	rcall	perform_operation
	rjmp	update
	
invert:
	;perform last requested operation (if any), then calculate reciprocal value of result
	rcall	perform_operation
	rcall	math_inv
	rjmp	update
	
sqrt:
	;perform last requested operation (if any), then calculate sqare root of result
	rcall	perform_operation
	rcall	math_sqrt
	rjmp	update
	
ce:
	;clear input
	ldi	inplen, 0
	ldi	decimals, 0
	andi	flags, ~(1<<InputNonzero)
	ori	flags, 1<<AnyInput | 1<<ManualInput
	rcall	clearnum2
	rjmp	update_nocopy
	
memclr:
	;clear memory
	rcall	clearmem
	rcall	io_updatesymbols
	rjmp	waitkey
	
memrcl:
	;recall memory
	andi	flags, ~(1<<ManualInput)
	ori	flags, 1<<AnyInput | 1<<InputNonzero
	ldi	inplen, 0
	ldi	decimals, 0
	rcall	copymemto2
	rjmp	update_nocopy
	
memadd:
	;add last result to memory
	sbrc	flags, AnyInput
	rcall	perform_operation
	rcall	copy1to3 ;backup last result to num3
	rjmp	memaddsub
	
memsub:
	;subtract last result from memory (by adding negative of last result)
	sbrc	flags, AnyInput
	rcall	perform_operation
	rcall	copy1to3 ;backup last result to num3
	ldi	temp, 1<<num1neg
	eor	flags, temp
	
memaddsub:
	;add num2 to num1, copy result to memory (and check if zero)
	;and restore last result from num3
	rcall	copymemto2
	rcall	math_add
	rcall	copy1tomem
	rcall	copy3to1 ;restore last result
	;set flag if memory is now zero ("empty")
	ori	flags2, 1<<MemZero
	ldi	ZL, mem
	ldi	count, DISPSIZE
memaddsub_zero:
	ld	temp, Z+
	cpi	temp, 0
	breq	PC+2
	andi	flags2, ~(1<<MemZero)
	dec	count
	brne	memaddsub_zero
	rjmp	update
	
	
	
percent:
	;ADD: multiply last result with (100 + entry)%
	;SUB: multiply last result with (100 - entry)%
	;MUL: multiply last result with (entry)%
	;DIV: divide last result by (entry)%
	;COPY: copy (entry)% == entry / 100 to result
	andi	flags, ~(1<<ShiftRounded) ;divide percentage by 100
	ldi	ZL, num2
	rcall	math_shiftright_round2
	ldi	ZL, num2
	rcall	math_shiftright_round2
	sbrc	flags2, OP_ADD
	rjmp	percent_addsub
	sbrc	flags2, OP_SUB
	rjmp	percent_addsub
	sbrc	flags2, OP_MUL
	rjmp	percent_mul
	sbrc	flags2, OP_DIV
	rjmp	percent_div
	rcall	copy2to1
	rjmp	percent_end
	
percent_addsub:
	rcall	copy1to3 ;backup value
	rcall	clearnum1
	ldi	ZL, num1+(DISPSIZE/2-1)
	ldi	temp, 0x01
	st	Z, temp
	ldi	temp, 1<<num2neg ;swap percentage's sign for subtraction
	sbrc	flags2, OP_SUB
	eor	flags, temp
	rcall	math_add ;add percentage to 1 (subtract from 1 if percentage is negative)
	rcall	copy3to2 ;restore value
	rcall	math_mul ;multiply percentage with value
	rjmp	percent_end

percent_mul:
	rcall	math_mul
	rjmp	percent_end

percent_div:
	rcall	math_div

percent_end:
	;clear number input helper registers (-> enter new number)
	rcall	perform_end
	rjmp	update

digit:
	;entered a digit
	cpi	inplen, DISPSIZE ;display is already full, can't enter any more
	brlo	PC+2
	rjmp	waitkey
	sbrc	flags, ManualInput ;clear num2 if this is the first digit
	rjmp	digit_process
	rcall	clearnum2
	andi	flags, ~(1<<InputNonzero)
	ldi	decimals, 0
digit_process:
	ori	flags, 1<<ManualInput | 1<<AnyInput
	cpi	temp, 0
	breq	PC+2
	ori	flags, 1<<InputNonzero
	sbrc	flags, InputNonzero ;don't add leading zeros to number length
	inc	inplen
	;append digit to existing number
	cpi	decimals, 1
	brsh	digit_decimal
	;integer digit: shift num2 left one place and put new digit as units
	ldi	ZL, num2
	push	temp
	rcall	math_shiftleft
	pop	temp
	ldi	ZL, num2+(DISPSIZE/2-1)
	ld	temp2, Z
	andi	temp2, 0xF0
	or	temp2, temp
	st	Z, temp2
	rjmp	digit_end
digit_decimal:
	;fractional digit: do not shift, append digit after last digit
	ldi	ZL, num2+(DISPSIZE/2-1) ;pointer to units digit
	mov	temp2, decimals
	inc	temp2
	lsr	temp2
	add	ZL, temp2
	ld	temp2, Z
	sbrc	decimals, 0
	swap	temp2
	andi	temp2, 0xF0
	or	temp2, temp
	sbrc	decimals, 0
	swap	temp2
	st	Z, temp2
	inc	decimals
digit_end:
	;calculate first displayed digit & length and update display
	ldi	ZL, num2
	rcall	numfirst
	cpi	first, DISPSIZE ;display at least 1 digit before decimal point
	brlo	PC+2
	ldi	first, DISPSIZE-1
	mov	length, inplen
	cpi	inplen, 0
	brne	PC+2
	inc	length
	rcall	io_update
	rjmp	waitkey
	
;===============================================================================

perform_operation:
	;perform requested operation with num1 and num2
	sbrs	flags, AnyInput ;do nothing if second operand is missing
	rjmp	perform_end
	sbrc	flags2, OP_ADD
	rjmp	perform_add
	sbrc	flags2, OP_SUB
	rjmp	perform_sub
	sbrc	flags2, OP_MUL
	rjmp	perform_mul
	sbrc	flags2, OP_DIV
	rjmp	perform_div
	rcall	copy2to1 ;copy num2 to num1
	rjmp	perform_end
perform_add:
	rcall	math_add ;add num2 to num1
	rjmp	perform_end

perform_sub:
	rcall	math_sub ;subtract num2 from num1
	rjmp	perform_end	
	
perform_mul:
	rcall	math_mul ;multiply num1 and num2
	rjmp	perform_end
	
perform_div:
	rcall	math_div ;divide num1 by num2

perform_end:
	;clear number input helper registers (-> enter new number)
	andi	flags, ~(1<<ManualInput | 1<<AnyInput | 1<<InputNonzero)
	andi	flags2, ~(1<<OP_ADD | 1<<OP_SUB | 1<<OP_MUL | 1<<OP_DIV)
	ldi	decimals, 0
	ldi	inplen, 0
	ret

;--------------------

clearnum1:
	;clear num1
	andi	flags, ~(1<<num1neg)
	ldi	ZL, num1
	rjmp	clearnum_all

clearnum2:
	;clear num2
	andi	flags, ~(1<<num2neg)
	ldi	ZL, num2
	rjmp	clearnum_all

clearnum3:
	;clear num3
	andi	flags, ~(1<<num3neg)
	ldi	ZL, num3
	rjmp	clearnum_all
	
clearmem:
	;clear memory
	andi	flags2, ~(1<<MemNeg)
	ori	flags2, 1<<MemZero
	ldi	ZL, mem
	rjmp	clearnum_all
	
clearnum_all:
	ldi	count, DISPSIZE
	ldi	temp2, 0
clearnum_loop:
	st	Z+, temp2
	dec	count
	brne	clearnum_loop
	ret
	
;--------------------

copy1to2:
	;copy num1 to num2 (with sign)
	ldi	ZL, num1+DISPSIZE
	ldi	count, DISPSIZE
copy1to2_loop:
	ld	temp2, -Z
	std	Z+(num2-num1), temp2
	dec	count
	brne	copy1to2_loop
	andi	flags, ~(1<<num2neg)
	sbrc	flags, num1neg
	ori	flags, 1<<num2neg
	ret

;--------------------

copy1to3:
	;copy num1 to num3 (with sign)
	ldi	count, DISPSIZE
	ldi	ZL, num1+DISPSIZE
copy1to3clear1_loop:
	ld	temp2, -Z
	std	Z+(num3-num1), temp2
	dec	count
	brne	copy1to3clear1_loop
	andi	flags, ~(1<<num3neg)
	sbrc	flags, num1neg
	ori	flags, 1<<num3neg
	ret

;--------------------

copy1tomem:
	;copy num1 to memory (with sign)
	ldi	count, DISPSIZE
	ldi	ZL, num1+DISPSIZE
copy1tomem_loop:
	ld	temp2, -Z
	std	Z+(mem-num1), temp2
	dec	count
	brne	copy1tomem_loop
	andi	flags2, ~(1<<MemNeg)
	sbrc	flags, num1neg
	ori	flags2, 1<<MemNeg
	ret

;--------------------

copy1tostart:
	;copy num1 to start value (no sign)
	ldi	count, DISPSIZE
	ldi	ZL, numstart
copy1tostart_loop:
	ldd	temp2, Z+(num1-numstart)
	st	Z+, temp2
	dec	count
	brne	copy1tostart_loop
	ret

;--------------------

copy1toiter:
	;copy num1 to iteration value (no sign)
	ldi	count, DISPSIZE
	ldi	ZL, numiter
copy1toiter_loop:
	ldd	temp2, Z+(num1-numiter)
	st	Z+, temp2
	dec	count
	brne	copy1toiter_loop
	ret

;--------------------

copy2to1:
	;copy num2 to num1 (with sign)
	ldi	ZL, num1
	ldi	count, DISPSIZE
copy2to1_loop:
	ldd	temp, Z+(num2-num1)
	st	Z+, temp
	dec	count
	brne	copy2to1_loop
	andi	flags, ~(1<<num1neg)
	sbrc	flags, num2neg
	ori	flags, 1<<num1neg
	ret

;--------------------

copy2to3:
	;copy num2 to num3 (with sign)
	ldi	ZL, num2+DISPSIZE
	ldi	count, DISPSIZE
copy2to3_loop:
	ld	temp, -Z
	std	Z+(num3-num2), temp
	dec	count
	brne	copy2to3_loop
	andi	flags, ~(1<<num3neg)
	sbrc	flags, num2neg
	ori	flags, 1<<num3neg
	ret

;--------------------

copy3to1:
	;copy num3 to num1 (with sign)
	ldi	ZL, num1
	ldi	count, DISPSIZE
copy3to1_loop:
	ldd	temp, Z+(num3-num1)
	st	Z+, temp
	dec	count
	brne	copy3to1_loop
	andi	flags, ~(1<<num1neg)
	sbrc	flags, num3neg
	ori	flags, 1<<num1neg
	ret

;--------------------

copy3to2:
	;copy num3 to num2 (with sign)
	ldi	ZL, num2
	ldi	count, DISPSIZE
copy3to2_loop:
	ldd	temp, Z+(num3-num2)
	st	Z+, temp
	dec	count
	brne	copy3to2_loop
	andi	flags, ~(1<<num2neg)
	sbrc	flags, num3neg
	ori	flags, 1<<num2neg
	ret

;--------------------

copymemto2:
	;copy memory to num2 (with sign)
	ldi	ZL, num2
	ldi	count, DISPSIZE
copymemto2_loop:
	ldd	temp, Z+(mem-num2)
	st	Z+, temp
	dec	count
	brne	copymemto2_loop
	andi	flags, ~(1<<num2neg)
	sbrc	flags2, MemNeg
	ori	flags, 1<<num2neg
	ret

;--------------------

copystartto1:
	;copy start value to num1 (no sign)
	ldi	ZL, numstart+DISPSIZE
	ldi	count, DISPSIZE
copystartto1_loop:
	ld	temp, -Z
	std	Z+(num1-numstart), temp
	dec	count
	brne	copystartto1_loop
	andi	flags, ~(1<<num1neg)
	ret

;--------------------

copyiterto2:
	;copy iteration value to num2 (no sign)
	ldi	ZL, numiter+DISPSIZE
	ldi	count, DISPSIZE
copyiterto2_loop:
	ld	temp, -Z
	std	Z+(num2-numiter), temp
	dec	count
	brne	copyiterto2_loop
	andi	flags, ~(1<<num2neg)
	ret

;--------------------

numfirst:
	;return position of first nonzero digit of the number starting at ZL
	;(if none, return DISPSIZE*2)
	ldi	first, 0
	ldi	count, DISPSIZE
numfirst_loop:
	ld	temp, Z+
	cpi	temp, 0x10
	brsh	numfirst_end
	inc	first
	cpi	temp, 0
	brne	numfirst_end
	inc	first
	dec	count
	brne	numfirst_loop
numfirst_end:
	ret
	
;--------------------

num2last:
	;return position of last nonzero digit after decimal point of num2
	;(if none, return position of units digit)
	ldi	last, DISPSIZE*2-1
	ldi	ZL, num2+DISPSIZE
num2last_loop:
	ld	temp, -Z
	swap	temp
	cpi	temp, 0x10
	brsh	num2last_end
	dec	last
	cpi	temp, 0
	brne	num2last_end
	dec	last
	cpi	last, DISPSIZE
	brsh	num2last_loop
num2last_end:
	ret
	
;-------------------
