;===============================================================================
;  input/output module: LED Display (up to 15 digits) and 4x6 keypad
;===============================================================================
;
;Ports:
;  PB0-7 = LED Segments a..g, dp
;  PD0   = Keypad Column 1
;  PD1   = Keypad Column 2
;  PD2   = Keypad Column 3
;  PD3   = Keypad Column 4
;  PD4-6 = LED Digit/Keypad Row Selection 0..2
;  PA0   = LED Digit/Keypad Row Selection 3
;
;Digit Selector:                                       _______
;                         (<= 8 digits: 74HC138, PA0 = Digit 0 = ENABLE[Pin6])
;                     4   ,------------------,          (>= 9 digits: 74HC154)
;  PA0 PD6 PD5 PD4 ===/===| Binary to Hex    |
;                         '-oooooooooooooooo-'
;                    _____  ||||||||||||||||
;                    Digit 0 .............. 15  <--  active low outputs
;
;LED Display:
;  Example for 8 digits: VQE12, 3x VQE14, VQE12-upside-down
;
;     0 15-(n-1) ... 14  15      <-- Digit 0, 15-(n-1)...15 (n = DISPSIZE)
;     |   |          |   |
;     R   R   ....   R   R       <-- Resistors (~ 1k)
;    /   /          /   /
;   / ,-/-o----//--/-o-/-o-------(+5V)
;  /BE|/BE|       /BE|/BE|
;  `|< `|<    ... `|< `|<        <-- PNP Transistors (e.g. BC327-40)
;    C|  C|         C|  C|
;     o---)----------)---)---,
;     |   |          |   |   |
;   ,---,---,--//--,---,---,---,   8
;   | - | 8.| .... | 8.| 8 |^+ |===/===[R]=== PB0..7, R = 8x ~33 Ohms
;   '---'---'--//--'---'---'---'
;
;   leftmost display: - = Segment g  (minus sign)
;  rightmost display: - = Segment a  (minus sign)
;  (optional)         ' = Segment b  (upper stroke of plus sign)
;                     , = Segment c  (lower stroke of plus sign)
;                     ^ = Segment dp (dot in upper left corner: memory =/= 0)
;Keypad (the layout can be changed below, search for "keytable:"):
;  ,---,---,---,---,
;  |1/x|sqr|   |   |---|>|--- Digit 9
;  +---+---+---+---+
;  |MR |MC |CE |AC |---|>|--- Digit 10
;  +---+---+---+---+
;  |M+ |M- |+/-| / |---|>|--- Digit 11
;  +---+---+---+---+
;  | 7 | 8 | 9 | * |---|>|--- Digit 12
;  +---+---+---+---+
;  | 4 | 5 | 6 | - |---|>|--- Digit 13
;  +---+---+---+---+
;  | 1 | 2 | 3 | + |---|>|--- Digit 14
;  +---+---+---+---+
;  | 0 | . | % | = |---|>|--- Digit 15
;  '---'---'---'---'
;    |   |   |   |
;   PD0 PD1 PD2 PD3

;number of display digits:
;  all functions available: 7 <= DISPSIZE <= 15
;  no sqrt and 1/x:         6 <= DISPSIZE <= 15
.equ	DISPSIZE = 8

.def	seldigit = R26 ;XL
.def	key = R27 ;XH

;===============================================================================
; interrupts
;===============================================================================

.org OVF0addr
	;timer 0 overflow interrupt
	push	temp
	push	temp2
	push	ZL
	push	ZH
	in	temp, SREG
	push	temp
	
	;TODO: check if new key routines work (old routines still in "../calculator_old/io_led_4x6.asm")
	;read keys in current row
	cpi	seldigit, 9
	brlo	io_ovf0_keyend
	ldi	ZL, LOW(keytable*2) ;set pointer to beginning of current row
	ldi	ZH, HIGH(keytable*2)
	mov	temp2, seldigit
	subi	temp2, 9
	lsl	temp2
	lsl	temp2
	add	ZL, temp2
	ldi	temp2, 0
	adc	ZH, temp2
	
	;check if a key in the current row is pressed
	in	temp, PIND
	ldi	temp2, 0xFF
	sbrs	temp, 0
	ldi	temp2, 0
	sbrs	temp, 1
	ldi	temp2, 1
	sbrs	temp, 2
	ldi	temp2, 2
	sbrs	temp, 3
	ldi	temp2, 3
	cpi	temp2, 0xFF ;no key in current row pressed if still 0xFF
	breq	io_ovf0_keyrelease
	;load keycode of pressed key from layout table
	add	ZL, temp2
	ldi	temp2, 0
	add	ZH, temp2
	lpm	key, Z
	rjmp	io_ovf0_keyend
	
io_ovf0_keyrelease:
	;check if a key in the current row was released
	ldi	temp2, 4
io_ovf0_keyrelease_loop:
	;read key codes in current row and compare to last keypress
	lpm	temp, Z+
	cp	key, temp
	breq	io_ovf0_keyreset
	dec	temp2
	brne	io_ovf0_keyrelease_loop
	rjmp	io_ovf0_keyend
io_ovf0_keyreset:
	ldi	key, 0xFF
io_ovf0_keyend:

	;increment digit pointer
	inc	seldigit
	andi	seldigit, 0x0F
	cpi	seldigit, 0
	breq	io_ovf0_incend
	cpi	seldigit, 15-(DISPSIZE-1)
	brsh	io_ovf0_incend
	ldi	seldigit, 15-(DISPSIZE-1)
io_ovf0_incend:

	;display off
	ldi	temp, 0xFF
	out	PORTB, temp
	
	;select new digit
	mov	temp, seldigit
	swap	temp
	ori	temp, 0x0F
	out	PORTD, temp
	ldi	temp, 0x00
	sbrc	seldigit, 3
	ldi	temp, 0x01
	out	PORTA, temp
	
	;output new digit
	ld	temp, X ;XL = seldigit
	out	PORTB, temp
	
	;return to main program
	pop	temp
	out	SREG, temp
	pop	ZH
	pop	ZL
	pop	temp2
	pop	temp
	reti
	
;--------------------
	
;keypad layout table, 4 cols x 7 rows (0xff = no function)
;this table can be changed if you with to use a different layout
;keytable: ;original layout
;	;   PD0         PD1         PD2          PD3
;	.db KEY_INV,    KEY_SQRT,   0xff,        0xff      ;--|>|-- digit 9
;	.db KEY_MEMRCL, KEY_MEMCLR, KEY_CE,      KEY_AC    ;--|>|-- digit 10
;	.db KEY_MEMADD, KEY_MEMSUB, KEY_NEG,     KEY_DIV   ;--|>|-- digit 11
;	.db 7,          8,          9,           KEY_MUL   ;--|>|-- digit 12
;	.db 4,          5,          6,           KEY_SUB   ;--|>|-- digit 13
;	.db 1,          2,          3,           KEY_ADD   ;--|>|-- digit 14
;	.db 0,          KEY_DOT,    KEY_PERCENT, KEY_EQUAL ;--|>|-- digit 15

keytable: ;alternative layout
	;   PD0         PD1         PD2          PD3
	.db KEY_INV,    KEY_SQRT,   0xff,        0xff      ;--|>|-- digit 9
	.db KEY_MEMRCL, KEY_MEMCLR, KEY_CE,      KEY_AC    ;--|>|-- digit 10
	.db KEY_MEMADD, KEY_MEMSUB, KEY_PERCENT, KEY_NEG   ;--|>|-- digit 11
	.db 7,          8,          9,           KEY_DIV   ;--|>|-- digit 12
	.db 4,          5,          6,           KEY_MUL   ;--|>|-- digit 13
	.db 1,          2,          3,           KEY_SUB   ;--|>|-- digit 14
	.db 0,          KEY_DOT,    KEY_EQUAL,   KEY_ADD   ;--|>|-- digit 15

;===============================================================================
; public functions (called from the main program)
;===============================================================================

io_init:
	;initialize I/O interface (called after power-on)
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0x00
	out	PORTA, temp
	ldi	temp, 0xFF
	out	DDRB, temp
	out	PORTB, temp
	ldi	temp, 0x70
	out	DDRD, temp
	ldi	temp, 0x0F
	out	PORTD, temp
	
	ldi	temp, 0x02 ;Clk/8
	out	TCCR0B, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	ldi	seldigit, 15-(DISPSIZE-1)
	ldi	key, 0xFF
	sei
	ret
	
;--------------------

io_update:
	;update display (called after number or operation changed)
	sbrs	flags, Error ;display "Error" if an error occured
	rjmp	io_update_noerror
	ldi	ZL, 15-(DISPSIZE-1)
	ldi	temp, 0x86 ;'E'
	st	Z+, temp
	ldi	temp, 0xAF ;'r'
	st	Z+, temp
	st	Z+, temp
	ldi	temp, 0xA3 ;'o'
	st	Z+, temp
	ldi	temp, 0xAF ;'r'
	st	Z+, temp
	ldi	count, DISPSIZE-5
	ldi	temp, 0xFF ;all segments off
io_update_error:
	st	Z+, temp
	dec	count
	brne	io_update_error
	ldi	ZL, 0
	st	Z, temp
	ret
io_update_noerror:
	;output spaces to right-align numbers
	push	length
	mov	temp, first
	lsr	temp
	ldi	ZL, num2
	add	ZL, temp
	mov	temp2, first
	ldi	YL, 15-(DISPSIZE-1)
	ldi	temp, 0xFF ;all segments off
	ldi	count, DISPSIZE
	sub	count, length
io_update_space:
	cpi	count, 0
	breq	io_update_num
	cpi	count, 1 ;use last digit before number (if any) for sign
	brne	io_update_space_nosign
	sbrc	flags, num2neg
	ldi	temp, 0xBF ;minus sign
io_update_space_nosign:
	st	Y+, temp
	dec	count
	rjmp	io_update_space
io_update_num:
	;output number digits (and decimal point after the units digit)
	ld	temp, Z
	sbrc	first, 0
	inc	ZL
	sbrs	first, 0
	swap	temp
	andi	temp, 0x0F
	push	ZL
	ldi	ZL, LOW(io_update_segments*2)
	ldi	ZH, HIGH(io_update_segments*2)
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm	temp, Z
	pop	ZL
	cpi	temp2, DISPSIZE-1
	brne	PC+2
	andi	temp, 0x7F ;dp enabled
	st	Y+, temp
	inc	temp2
	inc	first
	dec	length
	brne	io_update_num
	pop	length
	
io_updatesymbols:
	;output next operation, sign, and memory state (zero/nonzero)
	ldi	YL, 0
	ldi	temp, 0xFF
	sbrc	flags, AnyInput
	rjmp	io_update_sign
	sbrc	flags2, OP_ADD
	ldi	temp, 0xF8 ;display minus and both strokes of plus sign
	sbrc	flags2, OP_SUB
	ldi	temp, 0xFE ;display only minus sign
	sbrc	flags2, OP_MUL
	ldi	temp, 0xFD ;display only upper stroke of plus sign ("dot")
	sbrc	flags2, OP_DIV
	ldi	temp, 0xF9 ;display both strokes of plus sign ("colon")
io_update_sign:
	cpi	length, DISPSIZE
	brlo	io_update_sign_end
	sbrc	flags, num2neg
	andi	temp, 0xBF ;display left-hand minus sign (negative number)
io_update_sign_end:
	sbrs	flags2, MemZero
	andi	temp, 0x7F ;display "MEMORY" dot
	st	Y, temp
	
	ret
	
io_update_segments:
	;     0     1     2     3     4     5     6     7
	.db 0xc0, 0xf9, 0xa4, 0xb0, 0x99, 0x92, 0x82, 0xf8
	;     8     9
	.db 0x80, 0x90
	
;--------------------

io_getkey:
	;wait for keypress and return keycode
	cpi	key, 0xFF ;wait for release
	brne	io_getkey
io_getkey_keypress: ;wait for keypress
	cpi	key, 0xFF
	breq	io_getkey_keypress
	mov	temp, key
	ret
